/*
    Copyright (C) HWPORT.COM
    All rights reserved.
    Author: JAEHYUK CHO <mailto:minzkn@minzkn.com>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#if __GNUC__ < 3
# define calendar_builtin_expect(m_expression,m_value) m_expression
#else
# define calendar_builtin_expect(m_expression,m_value) __builtin_expect((long)(m_expression),(long)(m_value))
#endif
#define calendar_unlikely(m_expression) calendar_builtin_expect(m_expression,0)

static int is_leap_year(unsigned int s_year);
static unsigned int mdays(unsigned int s_year, unsigned int s_month);
static unsigned long days(unsigned int s_year, unsigned int s_month, unsigned int s_day);
static const char *week_label(unsigned int s_days);
static void *calendar_info(void *s_calendar_info_ptr, unsigned int s_year, unsigned int s_month, unsigned int s_day);
static void print_calendar(void *s_calendar_info_ptr);
static void print_days_info(unsigned int s_year, unsigned int s_month, unsigned int s_day, unsigned int s_hour, unsigned int s_min, unsigned int s_sec);

int main(void);

static int is_leap_year(unsigned int s_year)
{
    return(((((s_year % 4u) == 0u) && ((s_year % 100u) != 0u)) || ((s_year % 400u) == 0u)) ? 1 : 0);
}

static unsigned int mdays(unsigned int s_year, unsigned int s_month)
{
    static unsigned int cg_mday_table[][12] = {
        {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
        {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
    };

    if(calendar_unlikely(s_month <= 0u)) { return(0u); }

    return(cg_mday_table[is_leap_year(s_year)][s_month - 1u]);
}

static unsigned long days(unsigned int s_year, unsigned int s_month, unsigned int s_day)
{
    unsigned long s_days = (unsigned long)s_day;
    
    for(s_day = 1u;s_day < s_month;s_day++) { s_days += (unsigned long)mdays(s_year, s_day); }
 
    if(calendar_unlikely(s_year <= 0u)) { return(s_days); }

    --s_year;
    s_days += ((unsigned long)s_year) * 365ul;
    s_days += ((unsigned long)s_year) / 4ul;
    s_days += ((unsigned long)s_year) / 400ul;
    s_days -= ((unsigned long)s_year) / 100ul;
 
    return(s_days);
}

static const char *week_label(unsigned int s_days)
{
    static const char *cg_week_string[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

    return(cg_week_string[s_days % 7]);
}

static void *calendar_info(void *s_calendar_info_ptr, unsigned int s_year, unsigned int s_month, unsigned int s_day)
{
    unsigned char *s_calendar_info;
    unsigned int s_shift, s_count;

#if 0L
    s_calendar_info = (unsigned char *)s_calendar_info_ptr;
    for(s_count = 0u;s_count < (7u * 6u);s_count++) { s_calendar_info[s_count] = (unsigned char)0u; }
#else
    s_calendar_info = (unsigned char *)memset(s_calendar_info_ptr, 0, (size_t)(7 * 6));
#endif

    s_shift = (unsigned int)(days(s_year, s_month, 1u) % 7lu);
    for(s_count = 0u;s_count < mdays(s_year, s_month);s_count++) {
        s_calendar_info[s_shift + s_count] = (unsigned char)(s_count + 1u);
    }

    if(calendar_unlikely((s_day <= 0u) || (s_day > (7u * 6u)))) { return(s_calendar_info_ptr); }

    s_calendar_info[s_shift + s_day - 1u] |= (unsigned char)0x80u;

    return(s_calendar_info_ptr);
}

static void print_calendar(void *s_calendar_info_ptr)
{
    unsigned char *s_calendar_info = (unsigned char *)s_calendar_info_ptr;
    unsigned int s_shift, s_count;

    /* dump calendar info */
    for(s_count = 0u;s_count < 7u;s_count++) { (void)fprintf(stdout, "===="); }
    (void)fprintf(stdout, "\n");

    for(s_count = 0u;s_count < 7u;s_count++) { (void)fprintf(stdout, "%3s ", week_label(s_count)); }
    (void)fprintf(stdout, "\n");
    
    for(s_count = 0u;s_count < 7u;s_count++) { (void)fprintf(stdout, "----"); }
    (void)fprintf(stdout, "\n");

    for(s_shift = 0u;s_shift < (7u * 6u);s_shift += 7u) {
        for(s_count = 0u;(s_count < 7u) && (s_calendar_info[s_shift + s_count] == ((unsigned char)0u));s_count++);
        if(s_count == 7u) { /* is empty line */ break; }

        for(s_count = 0u;s_count < 7u;s_count++) {
            if(s_calendar_info[s_shift + s_count] == ((unsigned char)0u)) {
                (void)fprintf(stdout, "%3s ", " ");
            }
            else if((s_calendar_info[s_shift + s_count] & ((unsigned char)0x80u)) == ((unsigned char)0u)) {
                (void)fprintf(stdout, "%3u ", (unsigned int)s_calendar_info[s_shift + s_count]);
	    }
            else {
                (void)fprintf(stdout, "\x1b[1;46m%3u\x1b[0m ", (unsigned int)(s_calendar_info[s_shift + s_count] & 0x7f));
            }
        }
        (void)fprintf(stdout, "\n");
    }

    for(s_count = 0u;s_count < 7u;s_count++) { (void)fprintf(stdout, "===="); }
    (void)fprintf(stdout, "\n");
}

static void print_days_info(unsigned int s_year, unsigned int s_month, unsigned int s_day, unsigned int s_hour, unsigned int s_min, unsigned int s_sec)
{
    unsigned long s_days = days(s_year, s_month, s_day); 
    unsigned char s_calendar_info[7 * 6];

    (void)fprintf(stdout,
        "%u/%u/%u %02u:%02u:%02u %s (days=%lu)\n",
        s_year, s_month, s_day, s_hour, s_min, s_sec,
        week_label(s_days), s_days
    );

    print_calendar(calendar_info((void *)(&s_calendar_info[0]), s_year, s_month, s_day));
    
    (void)fprintf(stdout, "\n");
}

int main(void)
{
    time_t s_time;
    struct tm s_tm;

    (void)time((time_t *)(&s_time));
    (void)memcpy((void *)(&s_tm), (const void *)localtime((time_t *)(&s_time)), sizeof(s_tm));
    print_days_info(
        ((unsigned int)s_tm.tm_year) + 1900,
        ((unsigned int)s_tm.tm_mon) + 1,
        (unsigned int)s_tm.tm_mday,
        (unsigned int)s_tm.tm_hour,
        (unsigned int)s_tm.tm_min,
        (unsigned int)s_tm.tm_sec
    );

    print_days_info(2007u, 3u, 9u, 13u, 10u, 20u);
    print_days_info(2000u, 2u, 28u, 13u, 10u, 20u);
    print_days_info(2000u, 3u, 1u, 13u, 10u, 20u);

    return(EXIT_SUCCESS);
}

/* vim: set expandtab: */
/* End of source */
